@@ -13,9 +13,18 @@ import sys
 import cStringIO
 from commands import *
 
+
 class FiltersSet(object):
-    def __init__(self, name):
+    def __init__(self, name, filter_name_pretext="# Filter:", filter_desc_pretext="# Description:"):
+        """Represents a set of one or more filters
+
+        :param name: the filterset's name
+        :param filter_name_pretext: the text that is used to mark a filters name (as comment preceding the filter)
+        :param filter_desc_pretext: the text that is used to mark a filters description
+        """
         self.name = name
+        self.filter_name_pretext = filter_name_pretext
+        self.filter_desc_pretext = filter_desc_pretext
         self.requires = []
         self.filters = []
 
@@ -28,10 +37,10 @@ class FiltersSet(object):
 
     def __isdisabled(self, fcontent):
         """Tells if a filter is disabled or not
-
+        
         Simply checks if the filter is surrounded by a "if false" test.
 
-        :param name: the filter's name
+        :param fcontent: the filter's name
         """
         if not isinstance(fcontent, IfCommand):
             return False
@@ -48,13 +57,18 @@ class FiltersSet(object):
                 else:
                     self.require(f.arguments["capabilities"])
                 continue
-            if cpt - 1 >= len(parser.hash_comments):
-                name =  "Unnamed rule %d" % cpt
-            else:
-                name = parser.hash_comments[cpt - 1].replace("# Filter: ", "")
-            self.filters += [{"name" : name,
-                              "content" : f,
-                              "enabled" : not self.__isdisabled(f)}]
+
+            name = "Unnamed rule %d" % cpt
+            description = ""
+            for comment in f.hash_comments:
+                if comment.startswith(self.filter_name_pretext):
+                    name = comment.replace(self.filter_name_pretext, "")
+                if comment.startswith(self.filter_desc_pretext):
+                    description = comment.replace(self.filter_desc_pretext, "")
+            self.filters += [{"name": name,
+                              "description": description,
+                              "content": f,
+                              "enabled": not self.__isdisabled(f)}]
             cpt += 1
 
     def require(self, name):
@@ -62,7 +76,6 @@ class FiltersSet(object):
 
         :param name: the extension's name
         """
-
         name = name.strip('"')
         if not name in self.requires:
             self.requires += [name]
@@ -72,6 +85,8 @@ class FiltersSet(object):
 
         Called just before this object is going to be dumped.
         """
+        if not len(self.requires):
+            return None
         reqcmd = get_command_instance("require")
         reqcmd.check_next_arg("stringlist", self.requires)
         return reqcmd
@@ -85,7 +100,7 @@ class FiltersSet(object):
          * one or more actions
 
         A condition must be given as a 3-uple of the form::
-
+        
           (test's name, operator, value)
 
         An action must be given as a 2-uple of the form::
@@ -102,79 +117,25 @@ class FiltersSet(object):
         ifcontrol = get_command_instance("if")
         mtypeobj = get_command_instance(matchtype, ifcontrol)
         for c in conditions:
-            _cmd = None
-
-            if c[0] == "not":
-                _cmd = get_command_instance(c[0], ifcontrol)
-                c = c[1]
-
-            if c[0] in ["true", "false"]:
+            if c[0] in ("true", "false"):
                 cmd = get_command_instance(c[0], ifcontrol)
-            elif c in ["true", "false"]:
-                cmd = get_command_instance(c, ifcontrol)
             elif c[0] == "size":
                 cmd = get_command_instance("size", ifcontrol)
                 cmd.check_next_arg("tag", c[1])
                 cmd.check_next_arg("number", c[2])
-            elif c[0] == "envelope":
-                cmd = get_command_instance("envelope", ifcontrol)
-                for param in c[1:]:
-                    try:
-                        cmd.check_next_arg("tag", param)
-                    except:
-                        try:
-                            cmd.check_next_arg("string", param)
-                        except:
-                            try:
-                                cmd.check_next_arg("stringlist", param)
-                            except:
-                                pass
-
             else:
                 cmd = get_command_instance("header", ifcontrol)
-                if c[0] == "header":
-                    params = c[1:]
-                else:
-                    params = c
-
-                for param in params:
-                    try:
-                        cmd.check_next_arg("tag", param)
-                    except:
-                        try:
-                            cmd.check_next_arg("string", param)
-                        except:
-                            try:
-                                cmd.check_next_arg("stringlist", param)
-                            except:
-                                pass
-
-            if not _cmd == None:
-                _cmd.check_next_arg("test", cmd)
-                mtypeobj.check_next_arg("test", _cmd)
-            else:
-                mtypeobj.check_next_arg("test", cmd)
-
+                cmd.check_next_arg("tag", c[1])
+                cmd.check_next_arg("string", c[0])
+                cmd.check_next_arg("string", c[2])
+            mtypeobj.check_next_arg("test", cmd)
         ifcontrol.check_next_arg("test", mtypeobj)
 
         for actdef in actions:
-            if actdef in [ "discard", "keep", "stop"]:
-                action = get_command_instance(actdef, ifcontrol, False)
-            else:
-                #self.require(actdef[0])
-                action = get_command_instance(actdef[0], ifcontrol, False)
-                if len(actdef) > 1:
-                    for arg in actdef[1:]:
-                        if isinstance(arg, basestring) or isinstance(arg, str):
-                            try:
-                                action.check_next_arg("string", arg)
-                            except:
-                                action.check_next_arg("tag", arg)
-                        elif isinstance(arg, list):
-                            action.check_next_arg("stringlist", arg)
-                        elif isinstance(arg, int):
-                            action.check_next_arg("number", arg)
-
+            self.require(actdef[0])
+            action = get_command_instance(actdef[0], ifcontrol, False)
+            for arg in actdef[1:]:
+                action.check_next_arg("string", arg)
             ifcontrol.addchild(action)
         return ifcontrol
 
@@ -187,15 +148,16 @@ class FiltersSet(object):
         :param matchtype: "anyof" or "allof"
         """
         ifcontrol = self.__create_filter(conditions, actions, matchtype)
-        self.filters += [{"name" : name, "content" : ifcontrol, "enabled" : True}]
+        self.filters += [{"name": name, "content": ifcontrol, "enabled": True}]
 
     def updatefilter(self, oldname, newname, conditions, actions, matchtype="anyof"):
         """Update a specific filter
 
         Instead of removing and re-creating the filter, we update the
-        content in order to keep the original ordre between filters.
+        content in order to keep the original order between filters.
 
-        :param name: the filter's name
+        :param oldname: the filter's current name
+        :param newname: the filter's new name
         :param conditions: the list of conditions
         :param actions: the list of actions
         :param matchtype: "anyof" or "allof"
@@ -210,6 +172,29 @@ class FiltersSet(object):
                 return True
         return False
 
+    def replacefilter(self, oldname, sieve_filter, newname=None, description=None):
+        """replace a specific sieve_filter
+
+        Instead of removing and re-creating the sieve_filter, we update the
+        content in order to keep the original order between filters.
+
+        :param oldname: the sieve_filter's current name
+        :param newname: the sieve_filter's new name
+        :param sieve_filter: the sieve_filter object as get from FiltersSet.getfilter()
+        """
+        if newname is None:
+            newname = oldname
+        for f in self.filters:
+            if f["name"] == oldname:
+                f["name"] = newname
+                f["content"] = sieve_filter
+                if description is not None:
+                    f['description'] = description
+                if not f["enabled"]:
+                    return self.disablefilter(newname)
+                return True
+        return False
+
     def getfilter(self, name):
         """Search for a specific filter
 
@@ -309,10 +294,16 @@ class FiltersSet(object):
 
         Available for debugging purposes
         """
-
-        self.__gen_require_command().dump()
+        print "Dumping filters set %s\n" % self.name
+        cmd = self.__gen_require_command()
+        if cmd:
+            print "Dumping requirements"
+            cmd.dump()
+            print
 
         for f in self.filters:
+            print "Filter Name:%s" % f["name"]
+            print "Filter Description:%s" % f["description"]
             f["content"].dump()
 
     def tosieve(self, target=sys.stdout):
@@ -325,9 +316,14 @@ class FiltersSet(object):
 
         :param target: file pointer where the sieve syntax will be printed
         """
-        self.__gen_require_command().tosieve(target=target)
-        target.write("\n")
+        cmd = self.__gen_require_command()
+        if cmd:
+            cmd.tosieve(target=target)
+            target.write("\n")
         for f in self.filters:
+            print >> target, self.filter_name_pretext + f["name"]
+            if len(f["description"]) > 0:
+                print >> target, self.filter_desc_pretext + f["description"]
             f["content"].tosieve(target=target)
 
 
@@ -335,7 +331,7 @@ if __name__ == "__main__":
     fs = FiltersSet("test")
 
     fs.addfilter("rule1",
-                 [("Sender", ":is", "toto@toto.com"),],
-                 [("fileinto", "Toto"),])
+                 [("Sender", ":is", "toto@toto.com"), ],
+                 [("fileinto", "Toto"), ])
     fs.tosieve()
-
+    
