@@ -8,31 +8,36 @@ language used to filter emails.
 This implementation is based on RFC 5228 (http://tools.ietf.org/html/rfc5228)
 
 """
+import codecs
 import re
 import sys
 
 from commands import get_command_instance, UnknownCommand, BadArgument, BadValue
 
+
 class ParseError(Exception):
     """Generic parsing error"""
+
     def __init__(self, msg):
         self.msg = msg
 
     def __str__(self):
         return "parsing error: %s" % self.msg
 
+
 class Lexer(object):
     """
-    The lexical analysis part.
+    The lexical analysis part. 
 
     This class provides a simple way to define tokens (with patterns)
     to be detected.
 
     Patterns are provided into a list of 2-uple. Each 2-uple consists
     of a token name and an associated pattern, example:
-
+    
       [("left_bracket", r'\['),]
     """
+
     def __init__(self, definitions):
         self.definitions = definitions
         parts = []
@@ -67,12 +72,12 @@ class Lexer(object):
 
             m = self.regexp.match(text, self.pos)
             if m is None:
-                lineno = self.curlineno()
                 raise ParseError("unknown token %s" % text[self.pos:])
 
             self.pos = m.end()
             yield (m.lastgroup, m.group(m.lastgroup))
 
+
 class Parser(object):
     """The grammatical analysis part.
 
@@ -90,12 +95,12 @@ class Parser(object):
         ("comma", r','),
         ("hash_comment", r'#.*$'),
         ("bracket_comment", r'/\*[\s\S]*?\*/'),
-        ("multiline", r'text:[^$]*[\r\n]+\.$'),
+        ("multiline", r'text:[^$]*?[\r\n]+\.$'),
         ("string", r'"([^"\\]|\\.)*"'),
         ("identifier", r'[a-zA-Z_][\w]*'),
         ("tag", r':[a-zA-Z_][\w]*'),
         ("number", r'[0-9]+[KMGkmg]?'),
-        ]
+    ]
 
     def __init__(self, debug=False):
         self.debug = debug
@@ -105,7 +110,7 @@ class Parser(object):
         if not self.debug:
             return
         for m in msgs:
-            print(m)
+            print m
 
     def __reset_parser(self):
         """Reset parser's internal variables
@@ -147,10 +152,13 @@ class Parser(object):
                     if len(self.__curcommand.parent.children) >= 2 else None
             if prevcmd is None or prevcmd.name not in self.__curcommand.must_follow:
                 raise ParseError("the %s command must follow an %s command" % \
-                                     (self.__curcommand.name,
-                                      " or ".join(self.__curcommand.must_follow)))
+                                 (self.__curcommand.name,
+                                  " or ".join(self.__curcommand.must_follow)))
 
         if not self.__curcommand.parent:
+            #collect current amount of hash comments for later parsing into names and desciptions
+            self.__curcommand.hash_comments = self.hash_comments
+            self.hash_comments = []
             self.result += [self.__curcommand]
 
         if not onlyrecord:
@@ -175,7 +183,7 @@ class Parser(object):
 
         ctype = self.__curcommand.get_type()
         if ctype == "action" or \
-                (ctype == "control" and \
+                (ctype == "control" and
                      not self.__curcommand.accept_children):
             if testsemicolon:
                 self.__set_expected("semicolon")
@@ -201,7 +209,7 @@ class Parser(object):
     def __stringlist(self, ttype, tvalue):
         """Specific method to parse the 'string-list' type
 
-        Syntax:
+        Syntax: 
             string-list = "[" string *("," string) "]" / string
                             ; if there is only a single string, the brackets
                             ; are optional
@@ -224,7 +232,7 @@ class Parser(object):
 
         This method acts as an entry point for 'argument' parsing.
 
-        Syntax:
+        Syntax: 
             string-list / number / tag
 
         :param ttype: current token type
@@ -253,7 +261,7 @@ class Parser(object):
 
         Syntax:
             *argument [ test / test-list ]
-
+        
         :param ttype: current token type
         :param tvalue: current token value
         :return: False if an error is encountered, True otherwise
@@ -278,7 +286,7 @@ class Parser(object):
             return True
 
         if self.__argument(ttype, tvalue):
-            return self.__check_command_completion()
+            return self.__check_command_completion(testsemicolon=False)
 
         return False
 
@@ -307,25 +315,16 @@ class Parser(object):
 
             if ttype != "identifier":
                 return False
-
             command = get_command_instance(tvalue, self.__curcommand)
-
             if command.get_type() == "test":
                 raise ParseError("%s may not appear as a first command" % command.name)
-
             if command.get_type() == "control" and command.accept_children \
-                    and command.has_arguments():
+                and command.has_arguments():
                 self.__set_expected("identifier")
-
             if self.__curcommand is not None:
-                cur_parent = self.__curcommand
-                while not cur_parent.addchild(command):
-                    if hasattr(cur_parent, 'parent'):
-                        cur_parent = cur_parent.parent
-                    else:
-                        raise ParseError("%s unexpected after a %s" % \
-                                (tvalue, self.__curcommand.name))
-
+                if not self.__curcommand.addchild(command):
+                    raise ParseError("%s unexpected after a %s" % \
+                                     (tvalue, self.__curcommand.name))
             self.__curcommand = command
             self.__cstate = self.__arguments
 
@@ -351,7 +350,7 @@ class Parser(object):
     def parse(self, text):
         """The parser entry point.
 
-        Parse the provided text to check for its validity.
+        Parse the provided text to check for its validity. 
 
         On success, the parsing tree is available into the result
         attribute. It is a list of sievecommands.Command objects (see
@@ -367,7 +366,7 @@ class Parser(object):
         try:
             for ttype, tvalue in self.lexer.scan(text):
                 if ttype == "hash_comment":
-                    self.hash_comments += [tvalue]
+                    self.hash_comments += [tvalue.strip()]
                     continue
                 if ttype == "bracket_comment":
                     continue
@@ -375,25 +374,25 @@ class Parser(object):
                     if not ttype in self.__expected:
                         if self.lexer.pos < len(text):
                             msg = "%s found while %s expected near '%s'" \
-                                % (ttype, "|".join(self.__expected), text[self.lexer.pos])
+                                  % (ttype, "|".join(self.__expected), text[self.lexer.pos])
                         else:
                             msg = "%s found while %s expected at end of file" \
-                                % (ttype, "|".join(self.__expected))
+                                  % (ttype, "|".join(self.__expected))
                         raise ParseError(msg)
                     self.__expected = None
 
                 if not self.__command(ttype, tvalue):
                     msg = "unexpected token '%s' found near '%s'" \
-                        % (tvalue, text[self.lexer.pos])
+                          % (tvalue, text[self.lexer.pos])
                     raise ParseError(msg)
             if self.__opened_blocks:
                 self.__set_expected("right_cbracket")
             if self.__expected is not None:
                 raise ParseError("end of script reached while %s expected" % \
-                                     "|".join(self.__expected))
+                                 "|".join(self.__expected))
 
-        except (ParseError, UnknownCommand, BadArgument, BadValue) as err:
-            self.error = "line %d: %s" % (self.lexer.curlineno(), str(err))
+        except (ParseError, UnknownCommand, BadArgument, BadValue), e:
+            self.error = "line %d: %s" % (self.lexer.curlineno(), str(e))
             return False
         return True
 
@@ -405,7 +404,7 @@ class Parser(object):
         :param name: the pathname of the file to parse
         :return: True on success (no error detected), False otherwise
         """
-        fp = open(name)
+        fp = codecs.open(name, encoding='utf8')
         content = fp.read()
         fp.close()
         return self.parse(content)
@@ -418,6 +417,7 @@ class Parser(object):
         for r in self.result:
             r.dump(target=target)
 
+
 if __name__ == "__main__":
     from optparse import OptionParser
 
@@ -427,20 +427,26 @@ if __name__ == "__main__":
                   help="Activate verbose mode")
     op.add_option("-d", "--debug", action="store_true", default=False,
                   help="Activate debug traces")
+    op.add_option("--tosieve", action="store_true",
+                  help="Print parser results using sieve")
     options, args = op.parse_args()
 
     if not len(args):
-        print("Nothing to parse, exiting.")
+        print "Nothing to parse, exiting."
         sys.exit(0)
 
     for a in args:
         p = Parser(debug=options.debug)
-        print("Parsing file %s... " % (a))
+        print "Parsing file %s... " % a,
         if p.parse_file(a):
-            print("OK")
+            print "OK"
             if options.verbose:
                 p.dump()
+            if options.tosieve:
+                for r in p.result:
+                    r.tosieve()
             continue
-        print("ERROR")
-        print(p.error)
+        print "ERROR"
+        print p.error
 
+    
